module global_vars
    implicit none
    integer, parameter :: N = 1024
    double precision, dimension(0:N-1, 0:N-1) :: OUTopt, IN, W, OUTbase
end module global_vars

program openmp_fortran
  use omp_lib
  use global_vars

  implicit none

  integer :: i, j, ii, jj, k, argc, nths
  character(len=4) :: argv
  integer :: equal, res
  double precision :: ini, tend, tbase, topt

  k = 3
  argc = command_argument_count()
  if (argc > 0) then
     call get_command_argument(1, argv)
     read (argv,'(I10)') k
  end if
  print *,"Setting k=", k
  
  ! warm-up caches
  call init()

  ! Basic execution
  ini = omp_get_wtime()
  call base(k)
  tend = omp_get_wtime()
  tbase = (tend - ini)

  ! Optimized execution
  ini = omp_get_wtime()
  call opt(k)
  tend = omp_get_wtime()
  topt = (tend - ini)

  res = cmp(k)
  nths = omp_get_max_threads()
  if (.not. res) then
     print *, "NoOK!",k,nths,tbase,topt
  else
     print *, "OK!",k,nths,tbase,topt
     print *, "Times:",tbase,topt
  end if
contains

  subroutine init()
    integer :: i, j
    double precision :: x

    do i = 0, N-1
      do j = 0, N-1
        IN(i, j) = i+1.0 
        W(i, j) = j+1.0
      end do
    end do
  end subroutine init

  function cmp(k) result(equal)
    integer, intent(in) :: k
    integer :: i, j
    logical :: equal
    equal = .true.

    do i = k, N - k - 1 
      do j = k, N - k - 1
        if (OUTbase(i, j) /= OUTopt(i, j)) then
          equal = .false.
          exit
        end if
      end do
      if (.not. equal) exit
    end do
  end function cmp

  subroutine base(k)
    integer, intent(in) :: k
    integer :: i, j, ii, jj

    !$omp parallel
    !$omp single
    !$omp taskloop
    do i = k, N - k - 1 
      do j = k, N - k - 1
        OUTbase(i, j) = 0.0
        do ii = -k, k
          do jj = -k, k
            OUTbase(i, j) = OUTbase(i, j) + IN(i + ii, j + jj) * W(k + ii, k + jj)
          end do
        end do
      end do
    end do
    !$omp end taskloop
    !$omp end single
    !$omp end parallel
  end subroutine base

subroutine STMT(lb, ub, k, i, j)
    implicit none
    integer, intent(in) :: lb, ub, k, i, j
    integer :: ii, jj
    ! Presupone que OUT, IN, y W son matrices globales
    do ii = -k, k
        do jj = lb, ub
            OUTopt(i, j - jj) = OUTopt(i, j - jj) + IN(i + ii, j) * W(k + ii, k + jj)
        end do
    end do
end subroutine STMT

subroutine opt(k)
    integer, intent(in) :: k
    integer :: i, j
    ! Parallel region with OpenMP
    !$omp parallel
    !$omp single
    !$omp taskloop
    do i = k, N-k-1
        ! Primera parte del bucle
        do j = 0, 2*k-1
            OUTopt(i, j+k) = 0.0
            call STMT(-k, -k+j, k, i, j)
        end do
        ! Segunda parte del bucle
        do j = 2*k, N-2*k-1
            OUTopt(i, j+k) = 0.0
            call STMT(-k, k, k, i, j)
        end do
        ! Tercera parte del bucle
        do j = N-2*k, N-1
            call STMT(j-N+k+1, k, k, i, j)
        end do
    end do
    !$omp end taskloop
    !$omp end single
    !$omp end parallel
end subroutine opt

end program openmp_fortran
